#include "kf_interface.h"
#include <cstdint>
#include <unistd.h>


namespace Localization {
namespace Ego {

void KFInterface::InitFilterState(std::shared_ptr<BaseMeasure> &mea)
{
  initFlag_ = false;
  InitP();
  InitQ();
  // Init state
  if (mea != nullptr) {
    //1.位姿赋值
    if (initPos_ == false) {
      if ((mea->type_ == MeasurementType::RTK)) {
        std::shared_ptr<RtkMeasure> pRtk;
        pRtk = std::dynamic_pointer_cast<RtkMeasure>(mea);
        if (pRtk->velocityVeh_ > 0.0 && (pRtk->isRtkValid_ == true)) {
          state_.time = mea->stamp_;
          state_.q = pRtk->quatUTM_;
          state_.p = pRtk->posUTM_;
          state_.isRtkValid_ = pRtk->isRtkValid_;
          initPos_ = true;
          initAtt_ = true;
          std::cout<<"Egmo-InitFilterState():filter init att && pos success"<<std::endl;
        } else {
          std::cout<<"Egmo-InitFilterState():filter init att && pos false,rtk's velocityVeh_ is"<<
                   pRtk->velocityVeh_<<",rtk valid is:"<<pRtk->isRtkValid_<<std::endl;
        }
      } else {
        std::cout<<"Egmo-InitFilterState():filter try to init pos,but measure is not rtk"<<std::endl;
      }
    } else {
      //2.速度赋值
      if (mea->type_ == MeasurementType::WHEEL) {
        std::shared_ptr<VehMeasure> pVeh;
        pVeh = std::dynamic_pointer_cast<VehMeasure>(mea);
        state_.time = pVeh->GetStamp();
        state_.v = state_.q * pVeh->vm_;
        initVelo_ = true;
        std::cout<<"Egmo-InitFilterState():filter init vel success"<<std::endl;
      } else {
        std::cout<<std::setprecision(15)<<"Egmo-InitFilterState():filter try to init vel,but measure type is:"<<
          static_cast<int>(mea->type_)<<",stamp:"<<mea->stamp_<<std::endl;
      }
    }

    //3.校验所有初始化项
    if ((initPos_ == true) && (initAtt_ == true) && (initVelo_ == true)) {
      std::cout<<"Egmo-InitFilterState():filter init success,set computeFlag_ to true"<<std::endl;
      initFlag_ = true;
      computeFlag_ = true;
    } else {
      std::cerr<<"Egmo-InitFilterState():filter init false"<<",initPos_:"<<initPos_<<",initAtt_:"<<initAtt_<<",initVelo_:"<<initVelo_<<std::endl;
    }
  } else {
    std::cout<<"Egmo-InitFilterState() false, mea == nullptr"<<std::endl;
  }
}

void KFInterface::InitQ()
{
  Eigen::Matrix3d mI = Eigen::Matrix3d::Identity();
  mQ_ = Eigen::MatrixXd::Zero(StateNoiseIndex::NOISE_TOTAL, StateNoiseIndex::NOISE_TOTAL);
  mQ_(StateNoiseIndex::ACC_NOISE, StateNoiseIndex::ACC_NOISE) = accPropagateVar_;
  mQ_(StateNoiseIndex::ACC_NOISE + 1, StateNoiseIndex::ACC_NOISE + 1) = accPropagateVar_;
  mQ_(StateNoiseIndex::ACC_NOISE + 2, StateNoiseIndex::ACC_NOISE + 2) = accPropagateVar_;
  mQ_.block<3, 3>(StateNoiseIndex::AV_NOISE, StateNoiseIndex::AV_NOISE) = avPropagateVar_ * mI;

  mQ_.block<3, 3>(StateNoiseIndex::ACC_RANDOM_WALK, StateNoiseIndex::ACC_RANDOM_WALK) =
      imuAccRandomWalk_ * mI;
  // gyro bias random walk
    mQ_.block<3, 3>(StateNoiseIndex::GYRO_RANDOM_WALK, StateNoiseIndex::GYRO_RANDOM_WALK) =
        imuGyroRandomWalk_ * mI;
}
void KFInterface::InitP()
{
  Eigen::Matrix3d mI = Eigen::Matrix3d::Identity();
  state_.mP = Eigen::MatrixXd::Zero(StateIndex::STATE_TOTAL,StateIndex::STATE_TOTAL);
  state_.mP.block<3, 3>(StateIndex::R, StateIndex::R) = initRVar_ * mI;
  state_.mP.block<3, 3>(StateIndex::P, StateIndex::P) = initPVar_ * mI;
  state_.mP.block<3, 3>(StateIndex::V, StateIndex::V) = initVVar_ * mI;

  state_.mP(StateIndex::ACC, StateIndex::ACC) = initAccVar_;
  state_.mP(StateIndex::ACC + 1, StateIndex::ACC + 1) = initAccVar_;
  state_.mP(StateIndex::ACC + 2, StateIndex::ACC + 2) = initAccVar_;
  state_.mP.block<3, 3>(StateIndex::AV, StateIndex::AV) = initAvVar_ * mI;
  InitSensorParamVariance();
}

void KFInterface::InitSensorParamVariance()
{
  Eigen::Matrix3d mI = Eigen::Matrix3d::Identity();
  state_.mP.block<3, 3>(StateIndex::BA, StateIndex::BA) = initImuAccBiasVar_ * mI;
  state_.mP.block<3, 3>(StateIndex::BG, StateIndex::BG) = initImuGyroBiasVar_ * mI;

    state_.mP(StateIndex::WS, StateIndex::WS) = initWheelScaleVar_ ;
    state_.mP.block<3, 3>(StateIndex::IMU_INSTALL_ANGLE, StateIndex::IMU_INSTALL_ANGLE)
        = initImuInstallAngle_ * mI;
}

void KFInterface::PredictByimu(std::vector<std::shared_ptr<BaseMeasure>> &meaQue, State &state)
{
  for (auto pMea : meaQue) {
    if (pMea->type_==MeasurementType::WHEEL && pMea->isStatic_ == true) {
      return;
    }
    if (pMea->type_ == MeasurementType::IMU) {
      std::shared_ptr<ImuMeasure> pImu = std::dynamic_pointer_cast<ImuMeasure>(pMea);
      auto dT = pImu->stamp_ - state.time;
      if (dT < 1e-5) {
        continue;
      }
      if (dT < 0.0) {
        std::cout<<std::setprecision(15)<<"PredictByimu():imu's delta time < 0; time jump"<<dT<<"s"
                 <<",imu's time:"<<pImu->stamp_<<",state_'time:"<<state.time<<std::endl;
      } else if (dT > 1.0){
        std::cout<<std::setprecision(15)<<"PredictByimu():imu's delta time > 1.0; time blocked"<<dT<<"s"
                 <<",imu's time:"<<pImu->stamp_<<",state_'time:"<<state.time<<std::endl;
      }
      state.time = pImu->stamp_;
      Eigen::Vector3d acc0 = state.q * (pImu->acc_ + pImu->gravity_);
      Eigen::Vector3d omg{pImu->gyro_ * 0.5};
      Eigen::Quaterniond dQ{1.0, omg(0), omg(1), omg(2)};
      state.q = (state.q *dQ).normalized();
      Eigen::Vector3d acc1 = state.q * (pImu->acc_ + pImu->gravity_);
      state.acc = (acc0 + acc1) * 0.5;
      state.av = pImu->gyro_;
      state.v += state.acc * dT;
      state.p += state.acc * 0.5 * dT * dT + state.v * dT;
    }
  }
}

void KFInterface::Compute(std::shared_ptr<BaseMeasure> &ptrMea)
{
//  std::cout<<std::setprecision(15)<<"Compute():打印参与计算的观测，时间是："<<ptrMea->stamp_<<",类型是:"<<
//  ptrMea->typeName_[static_cast<int>(ptrMea->type_)]
//    <<",seq是："<<ptrM->seq_<<std::endl;
  State predictedState,updatedState;
  GetState(predictedState);
  GetState(updatedState);
  {
    //nan值判断
    static int nanCount_ = 0;
    if (nanCount_ >= 1000) {
      nanCount_ = 1001;
      std::cerr<<"nanCount_超过1000帧,KFInterface::ResetKF()"<<std::endl;
      ResetKF();
      return;
    }
    if(isnan(predictedState.mP(0,0))){
      nanCount_++;
      std::cout<<std::setprecision(15)<<"计算错误，出现nan值，return,时间戳是:"<<predictedState.time<<std::endl;
    }
  }
  if (!ptrMea->isStatic_){
    PredictToTime(ptrMea->stamp_, predictedState);
    {
      std::cout<<"预测：观测类型为："<<static_cast<int>(ptrMea->type_)<<"，打印一次状态"<<std::endl;
      predictedState.Print();
    }
    Update(ptrMea, predictedState,updatedState);
    {
      std::cout<<"观测：打印一次状态"<<std::endl;
      updatedState.Print();
    }
  } else {
    std::cout<<"Compute():由于静止,此次不计算"<<std::endl;
    updatedState.time = ptrMea->stamp_;
  }
  SetState(updatedState);
}

void KFInterface::PredictToTime(const double &time, State &state)
{
  double dT = time - state.time;

  if (dT > 1){
    std::cerr<<std::setprecision(15)<<"PredictToTime():measure delta delta time > 0.5;"
        <<"measure jump to future,dT is:"<<dT<<"s,return.wait for next compute and measure"<<std::endl;
    return;
  }
  State curState = state;

  while (dT >= 0.01) {//节省算力，改成20ms预测一次
      State nextState;
      PredictMean(curState, 0.01, nextState);
      Propagate(curState, 0.01, nextState);
      curState = nextState;
      dT -= 0.01;
    }
    State nextState = curState;
    if(dT > 1e-4) {
      PredictMean(curState, dT, nextState);
      Propagate(curState, dT, nextState);
    } else {
      std::cout<<std::setprecision(15)<<"补充预测失败,当前时间戳:"<<curState.time<<",dT is:"<<dT<<
        ",dT太小,跳过这次补充预测,直接进行观测"<<std::endl;
    }
    state = nextState;
}

void KFInterface::PredictMean(const State &currState, double dt, State &nextState)
{
  nextState.time = currState.time + dt;
  nextState.p = currState.p + currState.v * dt + currState.acc * 0.5 * dt * dt;
  nextState.v = currState.v + currState.acc * dt;
  Eigen::Vector3d omg{currState.av};
  omg = omg * dt / 2.0;
  Eigen::Quaterniond dQ{1.0, omg(0), omg(1), omg(2)};
  nextState.q = (currState.q * dQ).normalized();
  nextState.acc = currState.acc;
  nextState.av = currState.av;
  nextState.ba = currState.ba;
  nextState.bg = currState.bg;
  nextState.imuInstallAngle = currState.imuInstallAngle;
  nextState.ws = currState.ws;
}

void KFInterface::Propagate(const State &currState, double dt, State &nextState)
{
  Eigen::Matrix3d mI = Eigen::Matrix3d::Identity();
  Eigen::MatrixXd mF = Eigen::MatrixXd::Zero(StateIndex::STATE_TOTAL, StateIndex::STATE_TOTAL);

  mF.block<3, 3>(StateIndex::R, StateIndex::R) = mI - SkewSymmetric(currState.av) * dt;
  mF.block<3, 3>(StateIndex::R, StateIndex::AV) = dt * mI;
  mF.block<3, 3>(StateIndex::P, StateIndex::P) = mI;
  mF.block<3, 3>(StateIndex::P, StateIndex::V) = dt * mI;
  mF.block<3, 3>(StateIndex::P, StateIndex::ACC) = 0.5 * dt * dt * mI;
  mF.block<3, 3>(StateIndex::V, StateIndex::V) = mI;
  mF.block<3, 3>(StateIndex::V, StateIndex::ACC) = mI * dt;
  mF.block<3, 3>(StateIndex::ACC, StateIndex::ACC) = mI;
  mF.block<3, 3>(StateIndex::AV, StateIndex::AV) = mI;
  mF.block<3, 3>(StateIndex::BA, StateIndex::BA) = mI;
  mF.block<3, 3>(StateIndex::BG, StateIndex::BG) = mI;
  mF(StateIndex::WS, StateIndex::WS) = 1.0;
  mF.block<3, 3>(StateIndex::IMU_INSTALL_ANGLE, StateIndex::IMU_INSTALL_ANGLE) = mI;

  Eigen::MatrixXd mU = Eigen::MatrixXd::Zero(StateIndex::STATE_TOTAL, StateNoiseIndex::NOISE_TOTAL);
  mU.block<3, 3>(StateIndex::ACC, StateNoiseIndex::ACC_NOISE) = mI * dt;
  mU.block<3, 3>(StateIndex::AV, StateNoiseIndex::AV_NOISE) = mI * dt;
  mU.block<3, 3>(StateIndex::BA, StateNoiseIndex::ACC_RANDOM_WALK) = mI * dt;
  mU.block<3, 3>(StateIndex::BG, StateNoiseIndex::GYRO_RANDOM_WALK) = mI * dt;

  nextState.mP = mF * currState.mP * mF.transpose() + mU * mQ_ * mU.transpose();
}

void KFInterface::Update(std::shared_ptr<BaseMeasure> &ptrMea, State &predictedstate,State &updatedState)
{
  auto dT = ptrMea->stamp_ - predictedstate.time;
  if (dT < 0.0) {
    std::cerr<<std::setprecision(15)<<"Update():measure delta time < 0;measure block,predictedstate'time:"
    <<predictedstate.time<<",measure's time:"<<ptrMea->stamp_<<",dT is:"<<dT<<"s"<<std::endl;
  } else if (dT > 0.5){
    int type = static_cast<int>(ptrMea->type_);
    std::cerr<<std::setprecision(15)<<"Update():measure delta delta time > 0.5;measure jump to future, type is"
             <<ptrMea->typeName_[static_cast<int>(ptrMea->type_)]<<",predictedstate'time:"<<predictedstate.time<<",measure's time:"<<ptrMea->stamp_<<",dT is:"<<dT<<"s"<<std::endl;
  }
  if (ptrMea->stamp_ < predictedstate.time) {
    std::cout<<"Update():Update failed，measure stamp is smaller than predictedstate stamp; measure stamp:"
    <<ptrMea->stamp_<<",predictedstate stamp:"<<predictedstate.time<<std::endl;
    return;
  }
  ptrMea->Update(predictedstate,updatedState);
  updatedState.isImuValid_ = ptrMea->isImuValid_;
  updatedState.isVehValid_ = ptrMea->isVehValid_;
  updatedState.isRtkValid_ = ptrMea->isRtkValid_;
  updatedState.isVlocValid_ = ptrMea->isVlocValid_;

  if (ptrMea->type_ == MeasurementType::RTK) {
    std::cout<<"Update():尝试一次交叉校验"<<std::endl;
    if (CrossCheck(ptrMea,updatedState) == false) {
      static int resetCount = 0;
      static double lastResetTime = 0.0;
      double curResetTime = ptrMea->stamp_;
      if (curResetTime - lastResetTime > 10) {
        resetCount++;
        std::cerr<<"Update():CrossCheck fail,ResetKF(),count is:"<<resetCount<<std::endl;
        ResetKF();
      }
      lastResetTime = curResetTime;
    }
  }
}

bool KFInterface::CrossCheck(std::shared_ptr<BaseMeasure> &ptrMea,State &updatedState)
{
  bool res = true;
  State tempState = updatedState;
  if (ptrMea->isRtkValid_ == true) {
    std::shared_ptr<RtkMeasure> pRtk;
    pRtk = std::dynamic_pointer_cast<RtkMeasure>(ptrMea);


    Eigen::Matrix4d rtkPose = Eigen::Matrix4d::Identity();
    rtkPose.block<3, 3>(0, 0) = pRtk->quatUTM_.toRotationMatrix();
    rtkPose.block<3, 1>(0, 3) = pRtk->posUTM_;
    std::cout << std::setprecision(15) << "CrossCheck():时间戳,rtk:" << ptrMea->stamp_
              << ",state_:" << state_.time << std::endl;
    std::cout << "CrossCheck():rtk的attRPY:" << pRtk->attitude_.transpose() * RAD2DEG << std::endl;
    std::cout << "CrossCheck():rtk的pos:" << pRtk->posUTM_.transpose() << std::endl;
    Eigen::Matrix4d statePose = Eigen::Matrix4d::Identity();
    statePose.block<3, 3>(0, 0) = tempState.q.toRotationMatrix();
    statePose.block<3, 1>(0, 3) = tempState.p;
    Eigen::Vector3d stateYPR = RotYPR2EulerYPR(tempState.q.toRotationMatrix());
    std::cout << "CrossCheck():ego的attYPR:" << stateYPR.transpose() * RAD2DEG << std::endl;
    std::cout << "CrossCheck():ego的pos:" << tempState.p.transpose() << std::endl;
    Eigen::Matrix4d diffPose = rtkPose.inverse() * statePose;//转到rtk位姿下
    Eigen::Matrix3d diffRot = diffPose.block<3, 3>(0, 0);
    Eigen::Vector3d diffTrans = diffPose.block<3, 1>(0, 3);
    Eigen::Vector3d diffYPR = RotYPR2EulerYPR(diffRot);
    std::cout << "CrossCheck():ego的在rtk下的YPR" << diffYPR.transpose() * RAD2DEG << std::endl;
    std::cout << "CrossCheck():ego的在rtk下的pos" << diffTrans.transpose() << std::endl;

    //姿态校验
//    if (diffYPR.norm() > 5) {//100hz则是5度
//      std::cerr<<"CrossCheck():error,stamp:"<<ptrMea->stamp_<<",ego与rtk相对的diffYPR:"<<
//               diffYPR.transpose()*RAD2DEG<<",姿态差异太大，重置滤波器"<<std::endl;
//      res = false;
//    }
    //位置校验
    if (fabs(diffTrans.x()) > 100 || fabs(diffTrans.y()) > 5) {
      std::cerr << "CrossCheck():error,stamp:" << ptrMea->stamp_ << ",ego与rtk相对的的diff xyz:" <<
                diffTrans.transpose() << ",位置差异太大，重置滤波器" << std::endl;
      res = false;
    }
  }
  return res;
}

void KFInterface::SetState(State &tempState)
{
  std::unique_lock<std::mutex> lock(mutexState_);
  state_.time = tempState.time;
  state_.q = tempState.q;
  state_.p = tempState.p;
  state_.v = tempState.v;
  state_.acc = tempState.acc;
  state_.av = tempState.av;
  state_.ba = tempState.ba;
  state_.bg = tempState.bg;
  state_.ws = tempState.ws;
  state_.imuInstallAngle = tempState.imuInstallAngle;
  state_.mP = tempState.mP;
  state_.att = RotYPR2EulerYPR(state_.q.toRotationMatrix());
  state_.isImuValid_ = tempState.isImuValid_;
  state_.isVehValid_ = tempState.isVehValid_;
  state_.isRtkValid_ = tempState.isRtkValid_;
  state_.isVlocValid_ = tempState.isVlocValid_;
}
void KFInterface::ResetKF() {
  state_.Init();
  initPos_ = false;
  initAtt_ = false;
  initVelo_ = false;
  computeFlag_ = false;
}
void KFInterface::GetState(State &tempState) {
  std::unique_lock<std::mutex> lock(mutexState_);
  tempState.time = state_.time;
  tempState.q = state_.q;
  tempState.p = state_.p;
  tempState.v = state_.v;
  tempState.acc = state_.acc;
  tempState.av = state_.av;
  tempState.ba = state_.ba;
  tempState.bg = state_.bg;
  tempState.ws = state_.ws;
  tempState.imuInstallAngle = state_.imuInstallAngle;
  tempState.mP = state_.mP;
  tempState.att = state_.att;
  tempState.isImuValid_ = state_.isImuValid_;
  tempState.isVehValid_ = state_.isVehValid_;
  tempState.isRtkValid_ = state_.isRtkValid_;
  tempState.isVlocValid_ = state_.isVlocValid_;
}

double KFInterface::GetStateTime()
{
  std::unique_lock<std::mutex> lock(mutexState_);
  return state_.time;
}

bool KFInterface::GetComputeFlag()
{
  return computeFlag_;
}

bool KFInterface::GetInitFlag()
{
  return initFlag_;
}

} // namespace Ego
} // namespace Localization
